/**********************************************************************
**   CALIBRAT.C                                                      **
**   Example program using Interupt Driven Serial Port Routines      **
**   Written 2/24/94 by Milton Watts                                 **
**   QUARZDYNE, Inc.                                                 **
**                                                                   **
**   Downloads Pressure and Temperature Coefficients                 **
**   to Series I Transducer Interface                                **
**   Modified 4/25/96: Add variable baud rate option                 **
**                     Use '/' to indicate /e or /bnnnn options      **
**********************************************************************/


#include <time.h>   /* clock(), CLK_TIC */
#include <bios.h>   /* _bios_equiplist */
#include <string.h> /* strtok() */
#include <stdio.h>  /* printf() etc */

#include "serial.h"

int wait_for_message(int seconds);

#define MAXCHAR 1024

int main(int argc, char * argv[]) {
	FILE *fp;
	int port, nports, echo=0;
	char *ptr, buf[MAXCHAR];
	char address[]="01";
	int i,nfiles=0,c;
	char file[2][40];
	char usage[]=	"\nUsage: CALIBRAT COMn [#nn] [/e] [/bnnnn] pres.cff temp.cft\n"
						"     COMn indicates the serial port (n=1 or 2)\n"
						"     [#nn] is the logical device interface address (default is 01)\n"
						"     [/e] allows for RS485 interfaces which echo outgoing characters\n"
						"     [/bnnnn] allows for baud rate setting nnnn=300|600|1200|2400|4800|9600\n"
						"     pres.cff and temp.cft indicate the primary and secondary coefficient files\n"
						"     brackets indicate optional parameters\n"
	;
	unsigned baud = BAUD_1200;  // default baud rate

	if (argc < 2 || strnicmp (argv[1],"COM",3) != 0) {
		printf(usage);
		return 1;
	}
	port = argv[1][3]-'1';
	nports = ( _bios_equiplist() & 0x0E00 ) >> 9;
	if (port < 0 || port >=nports ) {
		printf("\nCommunications port %s not found\n",argv[1]);
		return 1;
	}

	for (i=2;i<argc;++i) {        /* read optional parameters */
		if ( *argv[i] == '#' && strlen(argv[i]) == 3 )  { // address specifier
			strcpy(address, argv[i]+1);
		} else if ( *argv[i] == '/' )  { // slash option
			switch ( *(argv[i]+1) ) {
			case 'E':
			case 'e':
				echo = 1;
				break;
			case 'B':
			case 'b':
				switch(*(argv[i]+2) ) {
				case '1':
					baud=_COM_1200;
					break;
				case '2':
					baud=_COM_2400;
					break;
				case '3':
					baud=_COM_300;
					break;
				case '4':
					baud=_COM_4800;
					break;
				case '6':
					baud=_COM_600;
					break;
				case '9':
					baud=_COM_9600;
					break;
				default:
					printf("\nInvalid baud rate specifier %s\n",argv[i]);
					printf(usage);
					exit(1);
				}
				break;
			}
		} else if (nfiles<2) {  // parameter must be a file specifier
			strcpy(file[nfiles++],argv[i]);
		} else {
			printf(usage);
			return 1;
		};
	};
	if (nfiles<1) {
		printf("At least 1 coefficient file must be specified\n");
		return 1;
	}

	/* first make sure the communications link is open */
	baud |= COMM_N81;  // settings = baud + No Parity + 8 Data bits + 1 stop	baud |= COMM_N81;
	if ( comm_init(port, baud, echo)==0 ) {
		printf("\nCommunications Port Cannot be initialized\n");
		printf("%s\n",error_message( comm_error(0) ) );
		return 1;
	}

	send_message(address,"ver");  /* check communications */
	if ( wait_for_message(2) == 0 ) {
		printf("\n\n\nNo response on COM%d address #%s\n", port+1, address);
		printf("Check cables, device address and COMn setting\n");
		comm_restore();
		return 1;
	}
	rcv_message(buf);
	printf(buf);

	for (i=0;i<2;++i) {
		if ( (fp=fopen(file[i],"r")) == NULL ) {
			printf("Unable to open coefficient file %s\n",file[i]);
			comm_restore();
			return 1;
		}

		ptr=buf+sprintf(buf,"cal%c={\n",i+'1'); /* build CAL command string */
		while ( ( c=fgetc(fp) )!=EOF) *ptr++=c; /* include file contents */
		strcpy(ptr,"}");                        /* complete the command */
		fclose(fp);

		send_message(address,buf);
		while ( message_sent() == 0);  /* wait for message to be sent */
		if (wait_for_message(10) == 0) {  /* wait up to 10 seconds for reply*/
			printf("No reply from interface downloading %s\n",file[i]);
			comm_restore();
			return(1);
		}
		while (wait_for_message(1) !=0) {
			rcv_message(buf);                  /* read each line of reply */
			printf(buf);
		}
		printf("\n");
	}
	send_message(address,"ew");
	if (wait_for_message(1) ==0) {
		printf("Unable to store Coefficients in permanent memory\n");
		comm_restore();
		return(1);
	}
	rcv_message(buf);
	printf("Coefficients stored in permanent memory\n%s",buf);
	comm_restore();
	return(0);
}


int wait_for_message(int seconds) {
	/* monitors message_rcvd() until either a message
		has been detected or until the number of seconds has elapsed.
		returns 1 if a message is detected, 0 otherwise.
	*/
	clock_t timeout;

	timeout= clock() + ( clock_t ) ( seconds * CLK_TCK ) ;
		 /* timeout in clock tics since program start */
	while ( message_rcvd(0) <=0 ) {  /* check for message first */
		if ( clock() > timeout ) return 0;
	}
	return 1;
}

