  ' Utility Functions for using Quartzdyne Coefficient files in Microsoft Excel
  '
  ' Written 10/30/03 by Milton Watts
  ' Modified 6/21/04 by Milton Watts
  '  - Added qdHexFindFt & qdHexFindFp based on existing qdFindFx functions
  ' Modified 7/11/2005 by Milton Watts - Release as separate text (.BAS) file for easy importing
  ' Modified 9/8/2008 by Milton Watts - Patch to fix bug in VAL() function
  '
  ' (c) 2003-2008 Quartzdyne, Inc.  All rights reserved.
  '
  ' The contents of this module may be used only in conjunction with Transducers and Coefficient Files
  ' supplied either directly or indirectly by Quartzdyne, Inc.
  ' Modifications are permitted as long as the original copyright notice remains in tact and all
  ' modifications are clearly notated and credited.
  ' This file may be copied and distributed as is, or as modified as long as the original copyright
  ' statment is included in the distribution.
  ' Modifications distributed to a third party remain the property of the party making the modifications,
  ' but are subject to the same terms as the original license.
  '
  ' To install these functions into a spreadsheet
  ' 1.) From the spreadsheet, open Microsoft Visual Basic Editor (Alt-F11)
  ' 2.) select "Import File" on the File menu
  ' 3.) navigate to this file (\Program Files\Quartzdyne\QDCalcFunctions.BAS) and click OK
  ' 4.) The functions may be accessed in the User Defined Function section of the Insert Function wizard.
  '
  
Public Function qdCfCalc(Fp As Double, Ft As Double, cfFile As Range) As Double
  ' Calculates pressure or temperature from counter readings
  ' using a standard Quartzdyne coefficient file.
  ' The file contents should be pasted into the spreadshet at any convenient location.
  ' cfFile should be set to entire range of cells that contain the coefficient file.
  '
  Dim xp As Double, xt As Double
  Dim i As Integer, j As Integer, n As Integer
  Dim Temp As Double
  
  ' Apply Prescaling to Fp and Ft
  xt = cfFile.Item(6) * (Ft - cfFile.Item(7))    ' xt = mt * (ft - ft0)
  xp = cfFile.Item(10) * (Fp - cfFile.Item(11))   ' xp = mp * (fp - fp0)
    
  ' n points to last coefficient
  n = (cfFile.Item(4).Value + 1) * (cfFile.Item(8).Value + 1) + 11 ' nCoef = (nt+1) * (np + 1)
  For i = 0 To cfFile.Item(8).Value  '  outer loop for np, xp polynomial
    qdCfCalc = qdCfCalc * xp                ' multiply by xp (0 first time through loop)
    Temp = 0
    For j = 0 To cfFile.Item(4).Value     ' inner loop for nt, xt polynomial
      Temp = xt * Temp + cfFile.Item(n)   ' process factored using coefficients from last to first
      n = n - 1
    Next j
    qdCfCalc = qdCfCalc + Temp  ' add temperature compensated pressure coefficient
  Next i
  
  '  Calculate output in user units using SPAN and ZERO
  n = i * j  ' find last coefficient again
  qdCfCalc = cfFile.Item(n + 12) * qdCfCalc + cfFile.Item(n + 13)  ' output = calc * SPAN + ZERO
End Function

Public Function qdFindFt(Target As Double, Fp As Double, cfFile As Range, Optional nSteps As Integer = 20, Optional Start As Double = 50000) As Double
  ' Iteratively searches for the Ft that results in the desired output when coefficients are applied.
  ' Fp is assumed constant, and should be calculated prior to calling this function.
  ' Ft is assumed to be in the range of 1Hz to 100kHz.  For applications where this is not the case,
  ' an alternate Start value should be specified
  ' This routine is NOT optimized.  It runs very slowly with large data sets.
  ' Accuracy can be sacrificed for speed by specifying a smaller value for nSteps
  '
  Dim Ft, Step As Double
  Dim err0 As Double, err1 As Double, err2 As Double
  Dim i As Integer
  
  Ft = Start
  Step = Start / 2 ' Allows a search from ( 0 + 2 ^ -nSteps ) to ( 2 * Start - 2 ^ -nSteps )
  err0 = qdCfCalc(Fp, Ft + 0, cfFile) - Target
  For i = 1 To nSteps ' binary search should yield precision of 1ppm in 20 loops ( 2^20 )
    err1 = qdCfCalc(Fp, Ft + Step, cfFile) - Target
    If err0 * err1 < 0 Then ' errors are different sign, we're heading in the correct direction
      Ft = Ft + Step
      err0 = err1
    Else
      err2 = qdCfCalc(Fp, Ft - Step, cfFile) - Target
      If err0 * err2 < 0 Or Abs(err1) > Abs(err2) Then ' sign change or smaller error
        Ft = Ft - Step
        err0 = err2
      Else
        Ft = Ft + Step
        err0 = err1
      End If
    End If
    Step = Step / 2
  Next i
  qdFindFt = Ft
  
End Function

Public Function qdFindFp(Target As Double, Ft As Double, cfFile As Range, Optional nSteps As Integer = 20, Optional Start As Double = 50000) As Double
  ' Iteratively searches for the Fp that results in the desired output when coefficients are applied.
  ' Ft is assumed constant, and should be calculated prior to calling this function.
  ' Fp is assumed to be in the range of 1Hz to 100kHz.  For applications where this is not the case,
  ' an alternate Start value should be specified
  ' This routine is NOT optimized.  It runs very slowly with large data sets.
  ' Accuracy can be sacrificed for speed by specifying a smaller value for nSteps
  '
  Dim Fp As Double, Step As Double
  Dim err0 As Double, err1 As Double, err2 As Double
  Dim i As Integer
  
  Fp = Start
  Step = Start / 2 ' Allows a search from ( 0 + 2 ^ -nSteps ) to ( 2 * Start - 2 ^ -nSteps )
  err0 = qdCfCalc(Fp + 0, Ft, cfFile) - Target
  For i = 1 To nSteps ' binary search should yield precision of 1ppm in 20 loops ( 2^20 )
    err1 = qdCfCalc(Fp + Step, Ft, cfFile) - Target
    If err0 * err1 < 0 Then ' errors are different sign, we're heading in the correct direction
      Fp = Fp + Step
      err0 = err1
    Else
      err2 = qdCfCalc(Fp - Step, Ft, cfFile) - Target
      If err0 * err2 < 0 Or Abs(err1) > Abs(err2) Then ' sign change or smaller error
        Fp = Fp - Step
        err0 = err2
      Else
        Fp = Fp + Step
        err0 = err1
      End If
    End If
    Step = Step / 2
  Next i
  qdFindFp = Fp
End Function

Public Function qdHexFileData(hexFile As Range) As String
  '
  ' Extract Hex Data from each line of the file range with NO error checking
  '
  ' hexFile is a Range of cells which includes the entire 16 lines of the hex coefficient file
  ' Return Value is the reduced Hex Data returned in a single string
  '
  Dim c As Variant
  
  For Each c In hexFile
    qdHexFileData = qdHexFileData & Mid(c, 10, 2 * Val("&H" & Mid(c, 2, 2)))
  Next
End Function

Public Function qdHexFileCheck(hexFile As Range) As String
  ' Extract Hex Data from each line of the file range WITH error checking
  '
  ' hexFile is a Range of cells which includes the entire 16 lines or the hex coefficient file
  ' Return Value is either the reduced Hex Data returned in a single string or an error message
  '
  Dim c As Variant
  Dim S As String
  Dim i As Integer, n As Integer
  Dim checkSum As Byte
  Dim nextByte As Long
       
  nextByte = 0
  
  ' Look at each Hex line individually
  For Each c In hexFile
    If Left(c, 1) <> ":" Then  ' Lines must begin with ":"
      qdHexFileCheck = "Error: Invalid Hex File Line <" & c & ">"
      Exit Function
    End If
    
    n = Val("&H" & Mid(c, 2, 2)) ' first field is length of data section in bytes
    
    If Mid(c, 8, 2) = "00" Then ' This is a data line
      If Val("&H" & Mid(c, 4, 4)) <> nextByte Then ' data must be complete and in order
        qdHexFileCheck = "Error: Hex data line is missing or out of order <" & c & ">"
        Exit Function
      End If
      
      ' Extract the data
      S = S & Mid(c, 10, 2 * Val("&H" & Mid(c, 2, 2)))
      nextByte = nextByte + n
    End If
      
    n = 2 * n + 11 ' expected length of line
    If Len(c) < n Then ' (ignore extra characters)
      qdHexFileCheck = "Error: Incomplete Hex File Line <" & c & ">"
      Exit Function
    End If
    checkSum = 0
    For i = 2 To n Step 2
      checkSum = (checkSum + Val("&H" & Mid(c, i, 2))) Mod 256
    Next i
    If checkSum <> 0 Then
      qdHexFileCheck = "Error: Invalid Hex Line Checksum (" & checkSum & ") <" & c & ">"
      Exit Function
    End If
  Next
  '
  ' Hex file succesfully parced
  ' Check for header and checksum
  If Left(S, 4) = "0D04" Then
    qdHexFileCheck = "Error: Little-Endian coefficient files not supported"
    Exit Function
  ElseIf Left(S, 4) <> "0D01" Then
    qdHexFileCheck = "Error: Unknown file type " & Left(S, 4)
    Exit Function
  End If
  
  If nextByte <> 256 Then
    qdHexFileCheck = "Error: Data set not complete (found " & nextByte & " bytes, expected 256)"
    Exit Function
  End If
  
  checkSum = 0
  For i = 1 To 511 Step 2
    checkSum = (checkSum + Val("&H" & Mid(S, i, 2))) Mod 256
  Next i
  If checkSum <> 0 Then
    qdHexFileCheck = "Error: Invalid internal checksum (" & checkSum & ")"
    Exit Function
  End If
  qdHexFileCheck = S
End Function

Public Function qdHexGetLong(hexData As String, Location As Integer) As Long
  ' returns data at byte Location in hexData interpreted as Long Integer
  '
  qdHexGetLong = Val("&H" & Mid(hexData, 2 * Location + 1, 8))
  '
  '  Bug in Val() function: if first 4 digits are "0000" they are ignored, and positive number may be interpreted as negative
  '
  If Mid(hexData, 2 * Location + 1, 4) = "0000" And qdHexGetLong < 0 Then
    qdHexGetLong = 2 ^ 16 + qdHexGetLong
  End If
End Function

Public Function qdHexGetInt(hexData As String, Location As Integer) As Integer
  qdHexGetInt = Val("&H" & Mid(hexData, 2 * Location + 1, 2))
End Function

Public Function qdHexGetFloat(hexData As String, Location As Integer) As Double
  Dim x As Long
  Dim sign As Integer, exp As Integer
  Dim mantissa As Double
  
  x = Val("&H" & Mid(hexData, 2 * Location + 1, 8))
  sign = Sgn(x)
  x = Abs(x)
  mantissa = (x Mod 2 ^ 23) / 2 ^ 23
  exp = Int(x / 2 ^ 23)
  
  qdHexGetFloat = sign * (1 + mantissa) * 2 ^ (exp - 127)
End Function

Public Function qdHexCalc(xp As Double, xt As Double, hexData As String, Optional calcT As Boolean = False, Optional altUnits As Boolean = False) As Double
  ' Calculates pressure from I2C Digital Transducer Readings
  ' using a Quartzdyne Hex coefficient file.
  ' The file contents should be pasted into the spreadshet at any convenient location.
  ' The data portion must then be extracted using either the qdHexFileData or qdHexFileCheck function.
  ' Presssure is calculated from the extracted data in the hexData string parameter.
  '
  Dim xpp As Double, xtp As Double
  Dim Z As Double, Temp As Double
  Dim i As Integer, j As Integer, n As Integer
  Dim np As Integer, nt As Integer
  '
  ' n points to last coefficient
  '
  xpp = xp * 2 ^ -24   ' Only prescale type 3 supported for digital coefficients
  xtp = xt * 2 ^ -24
  If calcT Then base = 140 Else base = 24
    
  n1 = qdHexGetInt(hexData, base + 2) ' Fit order in Xp
  n2 = qdHexGetInt(hexData, base + 3) ' Fit order in Xt
  n = 4 * (n1 + 1) * (n2 + 1) + base + 12 ' point to last coefficient
  Z = 0
  For i = 0 To n1       '  outer loop for np, xp polynomial
    Z = Z * xpp         ' multiply by xp (0 first time through loop)
    Temp = 0
    For j = 0 To n2     ' inner loop for nt, xt polynomial
      Temp = xtp * Temp ' process factored using coefficients from last to first
      Temp = Temp + qdHexGetLong(hexData, n)
      n = n - 4
    Next j
    Z = Z + Temp  ' add temperature compensated pressure coefficient
  Next i
  
  If (altUnits) Then
    Z = Z + qdHexGetLong(hexData, base + 12)          ' OFS2 placed just before coefficients
    qdHexCalc = Z * qdHexGetFloat(hexData, base + 8) ' alternate scale factor S2
  Else
    qdHexCalc = Z * qdHexGetFloat(hexData, base + 4) ' primary units scale factor S1
  End If
End Function

Public Function qdHexFindFt(Target As Double, Fp As Double, hexData As String, Optional nSteps As Integer = 20, Optional Start As Double = 30000000) As Double
  ' Iteratively searches for the Ft that results in the desired output when coefficients are applied.
  ' Fp is assumed constant, and should be calculated prior to calling this function.
  ' Ft is assumed to be in the range of 1Hz to 100kHz (600-60,000,000).
  ' For applications where this is not the case, an alternate Start value should be specified
  ' This routine is NOT optimized.  It runs very slowly with large data sets.
  ' Accuracy can be sacrificed for speed by specifying a smaller value for nSteps
  '
  Dim Ft, Step As Double
  Dim err0 As Double, err1 As Double, err2 As Double
  Dim i As Integer
  
  Ft = Start
  Step = Start / 2 ' Allows a search from ( 0 + 2 ^ -nSteps ) to ( 2 * Start - 2 ^ -nSteps )
  err0 = qdHexCalc(Fp, Ft + 0, hexData, True) - Target
  For i = 1 To nSteps ' binary search should yield precision of 1ppm in 20 loops ( 2^20 )
    err1 = qdHexCalc(Fp, Ft + Step, hexData, True) - Target
    If err0 * err1 < 0 Then ' errors are different sign, we're heading in the correct direction
      Ft = Ft + Step
      err0 = err1
    Else
      err2 = qdHexCalc(Fp, Ft - Step, hexData, True) - Target
      If err0 * err2 < 0 Or Abs(err1) > Abs(err2) Then ' sign change or smaller error
        Ft = Ft - Step
        err0 = err2
      Else
        Ft = Ft + Step
        err0 = err1
      End If
    End If
    Step = Step / 2
  Next i
  qdHexFindFt = Ft
  
End Function

Public Function qdHexFindFp(Target As Double, Ft As Double, hexData As String, Optional nSteps As Integer = 20, Optional Start As Double = 30000000) As Double
  ' Iteratively searches for the Fp that results in the desired output when coefficients are applied.
  ' Ft is assumed constant, and should be calculated prior to calling this function.
  ' Fp is assumed to be in the range of 1Hz to 100kHz (600-60,000,000).
  ' For applications where this is not the case, an alternate Start value should be specified.
  ' This routine is NOT optimized.  It runs very slowly with large data sets.
  ' Accuracy can be sacrificed for speed by specifying a smaller value for nSteps
  '
  Dim Fp As Double, Step As Double
  Dim err0 As Double, err1 As Double, err2 As Double
  Dim i As Integer
  
  Fp = Start
  Step = Start / 2 ' Allows a search from ( 0 + 2 ^ -nSteps ) to ( 2 * Start - 2 ^ -nSteps )
  err0 = qdHexCalc(Fp + 0, Ft, hexData) - Target
  For i = 1 To nSteps ' binary search should yield precision of 1ppm in 20 loops ( 2^20 )
    err1 = qdHexCalc(Fp + Step, Ft, hexData) - Target
    If err0 * err1 < 0 Then ' errors are different sign, we're heading in the correct direction
      Fp = Fp + Step
      err0 = err1
    Else
      err2 = qdHexCalc(Fp - Step, Ft, hexData) - Target
      If err0 * err2 < 0 Or Abs(err1) > Abs(err2) Then ' sign change or smaller error
        Fp = Fp - Step
        err0 = err2
      Else
        Fp = Fp + Step
        err0 = err1
      End If
    End If
    Step = Step / 2
  Next i
  qdHexFindFp = Fp
End Function

Sub testFindFt()
  Dim x As Double

  x = qdFindFt(28, 0, Worksheets("Freq Examples").Range("B1:B30"))
  MsgBox (x)
End Sub

Sub testqdCfCalc()
  Dim x As Double
  
  x = qdCfCalc(25000, 50000, Worksheets("Freq Examples").Range("A1:A30"))
  MsgBox (x)
End Sub

Sub testHex()
  Dim S As String
  Dim n As Long
  Dim f As Double
  Dim x As Double
  
  'S = qdHexFileData(Worksheets("RawHex").Range("A1:A17"))
  x = qdHexCalc(100000, 20000, Worksheets("RawHex").Range("A20"), False)
  ' f = qdhexGetFloat(S, 32)
  ' n = qdhexGetLong(S,40)
  ' n = qdhexGetInt(S,40)
  MsgBox (x)
End Sub



