/**********************************************************************
**   ROMREAD.C                                                       **
**   Example program using Interupt Driven Serial Port Routines      **
**   Written 4/25/97 by Milton Watts                                 **
**   QUARZDYNE, Inc.                                                 **
**                                                                   **
**   Queries a Series I for all ROMable data                         **
**   Data is stored to a file for use with the ROMWRITE program      **
**                                                                   **
**                                                                   **
**********************************************************************/


#include <time.h>   /* clock(), CLK_TIC */
#include <bios.h>   /* _bios_equiplist */
#include <string.h> /* strtok() */
#include <stdio.h>  /* printf() etc */

#include "serial.h"

int getlineNS(FILE *fp, char *line) ;
int getcmd(FILE *fp, char *cmd, int ver) ;
int getver(char *ptr) ;
int wait_for_message(int seconds);

#define MAXCHAR 1024

int main(int argc, char * argv[]) {
	FILE *fp1,*fp2;
	int port, nports, echo=0, ver;
	char *ptr, reply[MAXCHAR],cmd[MAXCHAR] ;
	char address[]="01";
	int i,nfiles=0,c;
	char cmdfile[40]="SICMD.TXT";
	char outfile[40]="SIDAT.TXT";
	char usage[]=	"\nUsage: ROMREAD COMn [#nn] [/e] [/bnnnn] [/fcmdfile] [outfile]"
						"     COMn indicates the serial port (n=1 or 2)\n"
						"     [#nn] is the logical device interface address (default is 01)\n"
						"     [/e] allows for RS485 interfaces which echo outgoing characters\n"
						"     [/bnnnn] allows for baud rate setting nnnn=300|600|1200|2400|4800|9600\n"
						"     [/fcmdfile] sets the command file path (default is SICMD.TXT)\n"
						"     [outfile] sets the ROM data output file (default is SIDAT.TXT)\n"
						"     brackets indicate optional parameters\n"
	;
	unsigned baud = BAUD_1200;  // default baud rate

	if (argc < 2 || strnicmp (argv[1],"COM",3) != 0) {
		printf(usage);
		return 1;
	}
	port = argv[1][3]-'1';
	nports = ( _bios_equiplist() & 0x0E00 ) >> 9;
	if (port < 0 || port >=nports ) {
		printf("\nCommunications port %s not found\n",argv[1]);
		return 1;
	}

	for (i=2;i<argc;++i) {        /* read optional parameters */
		if ( *argv[i] == '#' && strlen(argv[i]) == 3 )  { // address specifier
			strcpy(address, argv[i]+1);
		} else if ( *argv[i] == '/' )  { // slash option
			switch ( *(argv[i]+1) ) {
			case 'F':
			case 'f':
				strcpy(cmdfile,argv[i]+2);
				break;
			case 'E':
			case 'e':
				echo = 1;
				break;
			case 'B':
			case 'b':
				switch(*(argv[i]+2) ) {
				case '1':
					baud=_COM_1200;
					break;
				case '2':
					baud=_COM_2400;
					break;
				case '3':
					baud=_COM_300;
					break;
				case '4':
					baud=_COM_4800;
					break;
				case '6':
					baud=_COM_600;
					break;
				case '9':
					baud=_COM_9600;
					break;
				default:
					printf("\nInvalid baud rate specifier %s\n",argv[i]);
					printf(usage);
					exit(1);
				}
				break;
			}
		} else if (nfiles<1) {  // parameter must be output file specifier
			strcpy(outfile,argv[i]);
			nfiles++;
		} else {
			printf(usage);
			return 1;
		};
	};

	/* first make sure the communications link is open */
	baud |= COMM_N81;  // settings = baud + No Parity + 8 Data bits + 1 stop	baud |= COMM_N81;
	if ( comm_init(port, baud, echo)==0 ) {
		printf("\nCommunications Port Cannot be initialized\n");
		printf("%s\n",error_message( comm_error(0) ) );
		return 1;
	}

	send_message(address,"Ver");  /* check communications-get ROM version */
	if ( wait_for_message(2) == 0 ) {
		printf("\n\n\nNo response on COM%d address #%s\n", port+1, address);
		printf("Check cables, device address and COMn setting\n");
		comm_restore();
		return 1;
	}
	rcv_message(reply);
	printf(reply);
	ptr=strstr(reply,"Ver ");
	if (ptr!= NULL) ver=getver(ptr+4);
	else {
		printf("\nInvalid response on COM%d address #%s\n", port+1, address);
		printf("Check cables, device address and COMn setting\n");
		comm_restore();
		return 1;
	}

	if ( (fp1=fopen(cmdfile,"r")) == NULL ) {
		printf("Unable to open command description file <%s>\n",cmdfile);
		comm_restore();
		return 1;
	}
	if ( (fp2=fopen(outfile,"w")) == NULL ) {
		printf("Unable to open data output file <%s>\n",cmdfile);
		close(fp1);
		comm_restore();
		return 1;
	}

	fprintf(fp2,"; ROMREAD ROM Ver %d.%02d\n",ver/100,ver%100);  // header comments
	fprintf(fp2,"; Address #%s\n",address);

	while (getcmd(fp1,cmd,ver) ) {    // read command from command file
		send_message(address,cmd);  // query instrument
		if ( wait_for_message(10) == 0 ) {
			printf("\nNo response on COM%d address #%s to command\n", port+1, address);
			printf("%s\n",cmd);
			comm_restore();
			fclose(fp1);
			fclose(fp2);
			return 1;
		}
		rcv_message(reply);
		fprintf(fp2,"%s=%s",cmd,reply);
		printf("%s=%s",cmd,reply);
		if (strchr(reply,('{'))!= NULL) { // multiline command
			while (wait_for_message(1) !=0 && strchr(reply,('}')) == NULL) {
				rcv_message(reply);		 // receive remaining lines of reply
				fprintf(fp2,reply);
				printf(reply);
			}
		}
	}
	fprintf(fp2,"ew\n");  // force an eeprom write during restoration

	comm_restore();
	fclose(fp1);
	fclose(fp2);
	printf("Commands Stored in file: %s\n",outfile);
	return(0);
}

int getlineNS(FILE *fp, char *line) {
// gets one line of text from fp, trimming whitespace and comments
	int c, nchars=0;
	while ( ( c=getc(fp) ) != EOF) {
		switch(c) {
		case ';': case '!':
			while (  (c=getc(fp) ) != EOF && c != '\n');//plow thru comment
			if (nchars>0) {  // if we've got something, return it
				line[nchars]='\0';
				return nchars;
			}  // else check on next line
			break;
		case ' ':case'\t':
			break; // ignore white space
		case '\n':
			if (nchars>0) {  // if we've got something, return it
				line[nchars]='\0';
				return nchars;
			}  // else check on next line
			break;
		default:
			line[nchars++]=c;
		}
	}
	line[nchars]='\0';// return what we have
	return nchars;
}

int getcmd(FILE *fp, char *cmd, int ver) {
	char buf[MAXCHAR], *ptr;
	int ver1,ver2;

	while ( getlineNS(fp,buf) ) {
		ptr=buf;
		ver1=0,ver2=1000;  // version limits only apply to one line
		while (*ptr != '\0') {
			switch(*ptr++) {
			case '#':
				if (ver>=ver1 && ver<ver2) {
					while ( (*cmd++ = *ptr++ ) != '\0');  // copy command
					return(1);
				} else *ptr='\0'; // command not valid for this ver, try next
				break;
			case '+':
				ver1=getver(ptr);
				ptr+=4;
				break;
			case '-':
				ver2=getver(ptr);
				ptr+=4;
				break;
			default:
				// unknown operator
				printf("Syntax Error in command file\n %s\n",buf);
				return (0);
			} // get next operator
		}// line did not contain a valid command - try next
	} // EOF
	return (0);
}

int getver(char *ptr) {
	int ver;  // convert "n.nn" to integer between 1 and 999
	ver = 100 * (ptr[0]-'0')+ 10 * (ptr[2]-'0') +ptr[3]-'0';
	return ver;
}

int wait_for_message(int seconds) {
	/* monitors message_rcvd() until either a message
		has been detected or until the number of seconds has elapsed.
		returns 1 if a message is detected, 0 otherwise.
	*/
	clock_t timeout;

	timeout= clock() + ( clock_t ) ( seconds * CLK_TCK ) ;
		 /* timeout in clock tics since program start */
	while ( message_rcvd(0) <=0 ) {  /* check for message first */
		if ( clock() > timeout ) return 0;
	}
	return 1;
}

